snoise = """
   //-----------------------------------------------------------------------------
   // https://github.com/ashima/webgl-noise/blob/master/src/noise3D.glsl
   vec3 mod289(vec3 x) { return x - floor(x * (1.0 / 289.0)) * 289.0; }
   vec4 mod289(vec4 x) { return x - floor(x * (1.0 / 289.0)) * 289.0; }
   vec4 permute(vec4 x) { return mod289(((x*34.0)+1.0)*x); }
   vec4 taylorInvSqrt(vec4 r) { return 1.79284291400159 - 0.85373472095314 * r; }

   float snoise(vec3 v) {
     const vec2  C = vec2(1.0/6.0, 1.0/3.0) ;
     const vec4  D = vec4(0.0, 0.5, 1.0, 2.0);

     vec3 i  = floor(v + dot(v, C.yyy) );
     vec3 x0 =   v - i + dot(i, C.xxx) ;

     vec3 g = step(x0.yzx, x0.xyz);
     vec3 l = 1.0 - g;
     vec3 i1 = min( g.xyz, l.zxy );
     vec3 i2 = max( g.xyz, l.zxy );

     vec3 x1 = x0 - i1 + C.xxx;
     vec3 x2 = x0 - i2 + C.yyy;
     vec3 x3 = x0 - D.yyy;

     i = mod289(i);
     vec4 p = permute( permute( permute(
     i.z + vec4(0.0, i1.z, i2.z, 1.0 ))
     + i.y + vec4(0.0, i1.y, i2.y, 1.0 ))
     + i.x + vec4(0.0, i1.x, i2.x, 1.0 ));

     float n_ = 0.142857142857;
     vec3  ns = n_ * D.wyz - D.xzx;

     vec4 j = p - 49.0 * floor(p * ns.z * ns.z);

     vec4 x_ = floor(j * ns.z);
     vec4 y_ = floor(j - 7.0 * x_ );

     vec4 x = x_ *ns.x + ns.yyyy;
     vec4 y = y_ *ns.x + ns.yyyy;
     vec4 h = 1.0 - abs(x) - abs(y);

     vec4 b0 = vec4( x.xy, y.xy );
     vec4 b1 = vec4( x.zw, y.zw );

     vec4 s0 = floor(b0)*2.0 + 1.0;
     vec4 s1 = floor(b1)*2.0 + 1.0;
     vec4 sh = -step(h, vec4(0.0));

     vec4 a0 = b0.xzyw + s0.xzyw*sh.xxyy ;
     vec4 a1 = b1.xzyw + s1.xzyw*sh.zzww ;

     vec3 p0 = vec3(a0.xy,h.x);
     vec3 p1 = vec3(a0.zw,h.y);
     vec3 p2 = vec3(a1.xy,h.z);
     vec3 p3 = vec3(a1.zw,h.w);

     vec4 norm = taylorInvSqrt(vec4(dot(p0,p0), dot(p1,p1), dot(p2, p2), dot(p3,p3)));
     p0 *= norm.x;
     p1 *= norm.y;
     p2 *= norm.z;
     p3 *= norm.w;

     vec4 m = max(0.6 - vec4(dot(x0,x0), dot(x1,x1), dot(x2,x2), dot(x3,x3)), 0.0);
     m = m * m;
     return 42.0 * dot( m*m, vec4( dot(p0,x0), dot(p1,x1), dot(p2,x2), dot(p3,x3) ) );
   }
  """


class App


  constructor: ->
    @initScene()
    @initObjects()
    @animate()


  initScene: =>
    @renderer = new THREE.WebGLRenderer(precision: 'mediump', clearColor: 0x111111, clearAlpha: 1)
    if not @renderer.supportsFloatTextures() or not @renderer.supportsVertexTextures()
      console.log 'float texture:', @renderer.supportsFloatTextures()
      console.log 'vertex texture:', @renderer.supportsVertexTextures()
      Detector.addGetWebGLMessage()
      return

    @renderer.setSize(window.innerWidth, window.innerHeight)

    @camera = new THREE.PerspectiveCamera(50, window.innerWidth / window.innerHeight, 1, 5000)
    @camera.position.z = 700

    @scene = new THREE.Scene()

    @controls = new THREE.OrbitControls(@camera)
    @controls.autoRotate = true
    @controls.autoRotateSpeed *= 0.5

    container = document.getElementById('container')
    container.appendChild(@renderer.domElement)

    @params =
      numParticles: 4
      numColors: 10
      attractivePower: 5
      attractiveRadius: 370
      repulsivePower: 5
      repulsiveRadius: 50
      particleSize: 8
      maxForce: 3
      restart: =>
        @initData()
        @forceUpdatePass.uniforms.reset.value = 1
        @positionUpdatePass.uniforms.reset.value = 1
        @particles.material.uniforms.kind.value = @forceUpdatePass.uniforms.tKind.value

    @gui = new dat.GUI()
    @gui.add(@params, 'numColors', 1, 30).name('Num colors').step(1)
    @gui.add(@params, 'attractivePower', 0, 200).name('Attractive power').onChange (value) =>
      @forceUpdatePass.uniforms.attractivePower.value = value
    @gui.add(@params, 'attractiveRadius', 1, 500).name('Attractive radius').onChange (value) =>
      @forceUpdatePass.uniforms.attractiveRadius.value = value
    @gui.add(@params, 'repulsivePower', 0, 200).name('Repulsive power').onChange (value) =>
      @forceUpdatePass.uniforms.repulsivePower.value = value
    @gui.add(@params, 'repulsiveRadius', 1, 100).name('Repulsive radius').onChange (value) =>
      @forceUpdatePass.uniforms.repulsiveRadius.value = value
    @gui.add(@params, 'maxForce', 0.1, 10.0).name('Max force').onChange (value) =>
      @forceUpdatePass.uniforms.maxForce.value = value
    @gui.add(@params, 'particleSize', 1, 32).name('Particle size').onChange (value) =>
      @particles.material.uniforms.particleSize.value = value
    @gui.add(@params, 'restart').name('Restart')

    @stats = new Stats()
    container.appendChild(@stats.domElement)

    window.addEventListener 'resize', =>
      @camera.aspect = window.innerWidth / window.innerHeight
      @camera.updateProjectionMatrix()
      @renderer.setSize(window.innerWidth, window.innerHeight)


  initObjects: =>
    @dataWidth = 32 * (@params.numParticles|0)
    @dataHeight = 32
    options =
      magFilter: THREE.NearestFilter
      minFilter: THREE.NearestFilter
      format: THREE.RGBAFormat
      type: THREE.FloatType
      depthBuffer: false
      stencilBuffer: false
      generateMipmaps: false

    @forceUpdater = new THREE.EffectComposer(@renderer, new THREE.WebGLRenderTarget(@dataWidth, @dataHeight, options))
    @forceUpdatePass = new THREE.ShaderPass
      uniforms:
        reset: type: 'f', value: 1
        dataSize: type: 'v2', value: new THREE.Vector2(@dataWidth, @dataHeight)
        attractivePower: type: 'f', value: @params.attractivePower
        attractiveRadius: type: 'f', value: @params.attractiveRadius
        repulsivePower: type: 'f', value: @params.repulsivePower
        repulsiveRadius: type: 'f', value: @params.repulsiveRadius
        maxForce: type: 'f', value: @params.maxForce
        tInitial: type: 't', value: null
        tCurrent: type: 't', value: null
        tKind: type: 't', value: null
        tPosition: type: 't', value: null
      vertexShader: """
        varying vec2 vUv;
        void main() {
          gl_Position = projectionMatrix * modelViewMatrix * vec4(position, 1.);
          vUv = uv;
        }
        """
      fragmentShader: """
        #{snoise}
        uniform float reset;
        uniform vec2 dataSize;
        uniform float attractivePower;
        uniform float attractiveRadius;
        uniform float repulsivePower;
        uniform float repulsiveRadius;
        uniform float maxForce;
        uniform sampler2D tInitial;
        uniform sampler2D tCurrent;
        uniform sampler2D tKind;
        uniform sampler2D tPosition;
        varying vec2 vUv;
        float rand(vec2 co) {
          return fract(sin(dot(co.xy, vec2(12.9898, 78.233))) * 43758.5453);
        }
        void main() {
          if (reset > 0.) {
            gl_FragColor = texture2D(tInitial, vUv);
          } else {
            vec3 f = texture2D(tCurrent, vUv).xyz;
            vec3 p = texture2D(tPosition, vUv).xyz;
            vec3 k = texture2D(tKind, vUv).rgb;
            for (int y = 0; y < int(dataSize.y); ++y) {
              for (int x = 0; x < int(dataSize.x); ++x) {
                vec3 node = texture2D(tPosition, vec2(x, y) / dataSize).xyz;
                vec3 kind = texture2D(tKind, vec2(x, y) / dataSize).rgb;
                vec3 d = node.xyz - p.xyz;
                float dist = length(d);
                if (1. < dist && dist < repulsiveRadius) {
                  f -= d / pow(dist, 3.) * repulsivePower;
                } else if (repulsiveRadius <= dist && dist < attractiveRadius && k == kind) {
                  f += d / pow(dist, 3.) * attractivePower;
                }
                if (1. < dist && k == kind) {
                  f += normalize(texture2D(tCurrent, vec2(x, y) / dataSize).xyz) / dist / dist;
                }
              }
            }
            if (length(f) > maxForce) {
              f = normalize(f) * maxForce;
            }
            f.x += (rand(p.xy) - .5) * .3;
            f.y += (rand(p.yz) - .5) * .3;
            f.z += (rand(p.zx) - .5) * .3;
            gl_FragColor.xyz = f * 0.9;
          }
        }
        """
    @forceUpdatePass.textureID = 'tCurrent'
    @forceUpdater.addPass(@forceUpdatePass)

    @positionUpdater = new THREE.EffectComposer(@renderer, new THREE.WebGLRenderTarget(@dataWidth, @dataHeight, options))
    @positionUpdatePass = new THREE.ShaderPass
      uniforms:
        reset: type: 'f', value: 1
        tInitial: type: 't', value: null
        tCurrent: type: 't', value: null
        tForce: type: 't', value: null
      vertexShader: """
        varying vec2 vUv;
        void main() {
          gl_Position = projectionMatrix * modelViewMatrix * vec4(position, 1.);
          vUv = uv;
        }
        """
      fragmentShader: """
        uniform float reset;
        uniform sampler2D tInitial;
        uniform sampler2D tCurrent;
        uniform sampler2D tForce;
        varying vec2 vUv;
        void main() {
          vec4 p;
          if (reset > 0.) {
            gl_FragColor = texture2D(tInitial, vUv);
          } else {
            vec4 p = texture2D(tCurrent, vUv) + texture2D(tForce, vUv);
            if (p.x < -250.) {
              p.x += 500.;
            } else if (p.x > 250.) {
              p.x -= 500.;
            }
            if (p.y < -250.) {
              p.y += 500.;
            } else if (p.y > 250.) {
              p.y -= 500.;
            }
            if (p.z < -250.) {
              p.z += 500.;
            } if (p.z > 250.) {
              p.z -= 500.;
            }
            gl_FragColor = p;
          }
        }
        """
    @positionUpdatePass.textureID = 'tCurrent'
    @positionUpdater.addPass(@positionUpdatePass)
    @initData()

    geometry = new THREE.Geometry()
    for y in [0...@dataHeight]
      for x in [0...@dataWidth]
        geometry.vertices.push(new THREE.Vector3(x, y, 0))
    material = new THREE.ShaderMaterial
      depthTest: false
      depthWrite: false
      transparent: true
      blending: THREE.AdditiveBlending
      uniforms:
        data: type: 't', value: null
        dataSize: type: 'v2', value: new THREE.Vector2(@dataWidth, @dataHeight)
        map: type: 't', value: @generateNodeTex()
        kind: type: 't', value: @forceUpdatePass.uniforms.tKind.value
        particleSize: type: 'f', value: @params.particleSize
      vertexShader: """
        uniform sampler2D data;
        uniform vec2 dataSize;
        uniform sampler2D kind;
        uniform float particleSize;
        varying vec4 vColor;
        void main() {
          vec4 p = texture2D(data, position.xy / dataSize);
          p = modelViewMatrix * vec4(p.xyz, 1.);
          gl_Position = projectionMatrix * p;
          gl_PointSize = particleSize * -500. / p.z;
          vColor = texture2D(kind, position.xy / dataSize);
          vColor.a = smoothstep(-1000., -300., p.z);
        }
        """
      fragmentShader: """
        uniform sampler2D map;
        varying vec4 vColor;
        void main() {
          gl_FragColor = texture2D(map, gl_PointCoord) * vColor;
        }
        """
    @particles = new THREE.ParticleSystem(geometry, material)
    @scene.add(@particles)

    @scene.add(new THREE.Mesh(new THREE.CubeGeometry(500, 500, 500), new THREE.MeshBasicMaterial(color: 0xffffff, wireframe: true, transparent: true, opacity: 0.2)))


  initData: =>
    data = new Float32Array(3 * @dataWidth * @dataHeight)
    for i in [0...@dataWidth * @dataHeight]
      data[i * 3] = (Math.random() - 0.5) * 3
      data[i * 3 + 1] = (Math.random() - 0.5) * 3
      data[i * 3 + 2] = (Math.random() - 0.5) * 3
    initForce = new THREE.DataTexture(data, @dataWidth, @dataHeight, THREE.RGBFormat, THREE.FloatType)
    initForce.minFilter = THREE.NearestFilter
    initForce.magFilter = THREE.NearestFilter
    initForce.needsUpdate = true
    @forceUpdatePass.uniforms.tInitial.value?.dispose()
    @forceUpdatePass.uniforms.tInitial.value = initForce

    data = new Float32Array(3 * @dataWidth * @dataHeight)
    for i in [0...@dataWidth * @dataHeight]
      data[i * 3] = (Math.random() - 0.5) * 500
      data[i * 3 + 1] = (Math.random() - 0.5) * 500
      data[i * 3 + 2] = (Math.random() - 0.5) * 500
    initPosition = new THREE.DataTexture(data, @dataWidth, @dataHeight, THREE.RGBFormat, THREE.FloatType)
    initPosition.minFilter = THREE.NearestFilter
    initPosition.magFilter = THREE.NearestFilter
    initPosition.needsUpdate = true
    @positionUpdatePass.uniforms.tInitial.value?.dispose()
    @positionUpdatePass.uniforms.tInitial.value = initPosition

    data = new Uint8Array(3 * @dataWidth * @dataHeight)
    colors = []
    for i in [0..@params.numColors]
      c = new THREE.Color()
      c.setHSL(i / @params.numColors, 1.0, 0.5)
      colors.push(c)
    for i in [0...@dataWidth * @dataHeight]
      c = colors[(Math.random() * @params.numColors) | 0]
      data[i * 3] = c.r * 0xff
      data[i * 3 + 1] = c.g * 0xff
      data[i * 3 + 2] = c.b * 0xff
    kind = new THREE.DataTexture(data, @dataWidth, @dataHeight, THREE.RGBFormat)
    kind.minFilter = THREE.NearestFilter
    kind.magFilter = THREE.NearestFilter
    kind.needsUpdate = true
    @forceUpdatePass.uniforms.tKind.value?.dispose()
    @forceUpdatePass.uniforms.tKind.value = kind


  generateNodeTex: ->
    canvas = document.createElement('canvas')
    canvas.width = 32
    canvas.height = 32
    ctx = canvas.getContext('2d')
    ctx.clearRect(0, 0, 32, 32)
    ctx.beginPath()
    ctx.arc(16, 16, 15, 0, Math.PI * 2, false)
    ctx.fillStyle = 'white'
    ctx.fill()
    tex = new THREE.Texture(canvas)
    tex.needsUpdate = true
    return tex


  swap: (target) =>
    tmp = target.renderTarget1
    target.renderTarget1 = target.renderTarget2
    target.renderTarget2 = tmp


  animate: =>
    requestAnimationFrame(@animate)

    @forceUpdatePass.uniforms.tPosition.value = @positionUpdater.renderTarget2
    @forceUpdater.render()
    @positionUpdatePass.uniforms.tForce.value = @forceUpdater.renderTarget1
    @positionUpdater.render()
    @particles.material.uniforms.data.value = @positionUpdater.renderTarget1

    @swap(@forceUpdater)
    @swap(@positionUpdater)

    @controls.update()
    @renderer.render(@scene, @camera)

    @forceUpdatePass.uniforms.reset.value = 0
    @positionUpdatePass.uniforms.reset.value = 0

    @stats.update()


if Detector.webgl
  new App()
else
  Detector.addGetWebGLMessage()
